﻿local ffi = require 'ffi'
ffi.cdef [[
typedef unsigned long uLong;
typedef unsigned int uInt;
typedef unsigned char Bytef;

typedef void* voidpf;
typedef voidpf (*alloc_func)(voidpf opaque,uInt items,uInt size);
typedef void   (*free_func) (voidpf opaque, voidpf address);

typedef struct z_stream_s {
  const Bytef *next_in;     /* next input byte */
  uInt  avail_in;  /* number of bytes available at next_in */
  uLong total_in;  /* total number of input bytes read so far */

  Bytef *next_out; /* next output byte should be put there */
  uInt  avail_out; /* remaining free space at next_out */
  uLong total_out; /* total number of bytes output so far */

  const char *msg; /* last error message, NULL if no error */
  struct internal_state *state; /* not visible by applications */

  alloc_func zalloc; /* used to allocate the internal state */
  free_func  zfree;  /* used to free the internal state */
  voidpf opaque; // private data object passed to zalloc and zfree

  int data_type; // best guess about the data type: binary or text
  uLong  adler; /* adler32 value of the uncompressed data */
  uLong  reserved; /* reserved for future use */
} z_stream, *z_streamp;
typedef struct { z_stream s; } z_stream2; // pour ffi.metatype

int deflateInit_ (z_streamp strm, int level,
  const char *version, int stream_size);
int deflate (z_streamp strm, int flush);
int deflateEnd (z_streamp strm);
int inflateInit_ (z_streamp strm,
  const char *version, int stream_size);
int inflate (z_streamp strm, int flush);
int inflateEnd (z_streamp strm);
]]
local zlib = ffi.load("z")
local flush_code = {}
for k,v in ipairs{"no", "partial", "sync", "full",
    "finish", "block", "trees"} do
  flush_code[v] = k-1
end

local function deflate_common(strm, src, fct, flush, end_result)
  local res, dst, bufsize = -1, {}, 1024
  local dstbuf = ffi.new('Bytef[?]', bufsize)
  strm.next_in = src
  strm.avail_in = #src
  while strm.avail_in > 0 or res < end_result do
    strm.next_out = dstbuf
    strm.avail_out = bufsize
    res = fct(strm, flush)
    assert(res >= 0)
    dst[#dst+1] = ffi.string(dstbuf, bufsize - strm.avail_out)
  end
  return table.concat(dst)
end

local deflmeta = {
  add = function(obj, data)
    return deflate_common(obj, data, zlib.deflate, 0, 0)
  end,
  flush = function(obj, mode)
    mode = mode or "finish"
    return deflate_common(obj, "", zlib.deflate,
      flush_code[mode], mode == "finish" and 1 or 0)
  end,
  __tostring = function(obj) return "ZLib deflater" end,
  __gc = zlib.deflateEnd,
}
deflmeta.__index = deflmeta
local defltype = ffi.metatype("z_stream", deflmeta)

local function alloc_init(strm)
  local alloc_table = {}
  local total_alloc = 0
  function strm.zalloc(opaque, items, size)
    local block = ffi.new("uint8_t[?]", items*size)
    local address = tonumber(ffi.cast('ptrdiff_t', block))
    alloc_table[address] = { block, items, size }
    total_alloc = total_alloc + items * size
    print(total_alloc.." octets alloués")
    return block
  end
  function strm.zfree(opaque, address)
    address = tonumber(ffi.cast('ptrdiff_t', address))
    local t = alloc_table[address]
    total_alloc = total_alloc - t[2] * t[3]
    print(total_alloc.." octets alloués")
    alloc_table[address] = nil
  end
end

local function deflater(level)
  local strm = defltype()
  alloc_init(strm)
  assert(zlib.deflateInit_(strm, level or -1, "1.2.6",
    ffi.sizeof(strm))==0)
  return strm
end

local inflmeta = {
  add = function(obj, data)
    return deflate_common(obj.s, data, zlib.inflate, 0, 0)
  end,
  flush = function(obj, mode)
    mode = mode or "finish"
    return deflate_common(obj.s, "", zlib.inflate,
      flush_code[mode], mode == "finish" and 1 or 0)
  end,
  __tostring = function(obj) return "ZLib inflater" end,
  __gc = function(obj) zlib.inflateEnd(obj.s) end,
}
inflmeta.__index = inflmeta
local infltype = ffi.metatype("z_stream2", inflmeta)

local function inflater()
  local strm = infltype()
  alloc_init(strm.s)
  assert(zlib.inflateInit_(strm.s, "1.2.6", ffi.sizeof(strm))==0)
  return strm
end

-- Test
local orig = ("Hello World!\n"):rep(10000)
local defl = deflater(9)
local compr = defl:add(orig)..defl:flush()
print(#orig, #compr) --> 130000  293
local infl = inflater()
local ucompr = infl:add(compr)..infl:flush()
assert(ucompr == orig)
